@@ -1,4 +1,4 @@
-/********************************************************************** 
+/***********************************************************************
  Freeciv - Copyright (C) 1996 - A Kjeldberg, L Gregersen, P Unold
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -17,6 +17,8 @@
 
 #include "colors_g.h"
 
-extern GdkColor *colors_standard[COLOR_STD_LAST];
+struct color {
+  GdkRGBA color;
+};
 
-#endif  /* FC__COLORS_H */
+#endif /* FC__COLORS_H */
