@@ -1,4 +1,4 @@
-/********************************************************************** 
+/***********************************************************************
  Freeciv - Copyright (C) 2005-2007 - The Freeciv Project
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -12,30 +12,30 @@
 ***********************************************************************/
 
 #ifdef HAVE_CONFIG_H
-#include <config.h>
+#include <fc_config.h>
 #endif
 
-#include <assert.h>
-#include <log.h>
 #include <stdarg.h>
 #include <string.h>
 
+/* utility */
+#include "log.h"
+
 /* common */
 #include "government.h"
 #include "improvement.h"
+#include "research.h"
 #include "tech.h"
 
 /* client */
 #include "client_main.h"
+#include "options.h"
 #include "tilespec.h"
+#include "reqtree.h"
 
 #include "colors_g.h"
 #include "sprite_g.h"
 
-#include "reqtree.h"
-#include "tilespec.h"
-#include "options.h"
-
 /*
  * Hierarchical directed draph drawing for Freeciv's technology tree
  *
@@ -61,7 +61,7 @@
 
 
 /****************************************************************************
-  This structure desribes a node in a technology tree diagram.
+  This structure describes a node in a technology tree diagram.
   A node can by dummy or real. Real node describes a technology.
 ****************************************************************************/
 struct tree_node {
@@ -115,11 +115,14 @@ enum reqtree_edge_type {
   REQTREE_GOAL_EDGE     /* Dest node is part of goal "future visited" */
 };
 
-/*************************************************************************
+/*********************************************************************//**
   Add requirement edge to node and provide edge to req
 *************************************************************************/
 static void add_requirement(struct tree_node *node, struct tree_node *req)
 {
+  fc_assert_ret(node != NULL);
+  fc_assert_ret(req != NULL);
+
   node->require =
       fc_realloc(node->require,
 		 sizeof(*node->require) * (node->nrequire + 1));
@@ -133,7 +136,7 @@ static void add_requirement(struct tree_node *node, struct tree_node *req)
   req->nprovide++;
 }
 
-/*************************************************************************
+/*********************************************************************//**
   Allocate and initialize new tree node
 *************************************************************************/
 static struct tree_node *new_tree_node(void)
@@ -149,37 +152,38 @@ static struct tree_node *new_tree_node(void)
   return node;
 }
 
-/****************************************************************************
+/*********************************************************************//**
   Return minimum size of the rectangle in pixels on the diagram which
   corresponds to the given node
-****************************************************************************/
+*************************************************************************/
 static void node_rectangle_minimum_size(struct tree_node *node,
-					int *width, int *height)
+                                        int *width, int *height)
 {
   int max_icon_height; /* maximal height of icons below the text */
   int icons_width_sum; /* sum of icons width plus space between them */
   struct sprite* sprite;
   int swidth, sheight;
-  
+
   if (node->is_dummy) {
     /* Dummy node is a straight line */
     *width = *height = 1;
   } else {
     get_text_size(width, height, FONT_REQTREE_TEXT,
-		  advance_name_for_player(client.conn.playing, node->tech));
+                  research_advance_name_translation
+                      (research_get(client_player()), node->tech));
     *width += 2;
     *height += 8;
-    
+
     max_icon_height = 0;
     icons_width_sum = 5;
-    
-    if (reqtree_show_icons) {
+
+    if (gui_options.reqtree_show_icons) {
       /* units */
       unit_type_iterate(unit) {
         if (advance_number(unit->require_advance) != node->tech) {
           continue;
         }
-        sprite = get_unittype_sprite(tileset, unit);
+        sprite = get_unittype_sprite(tileset, unit, direction8_invalid());
         get_sprite_dimensions(sprite, &swidth, &sheight);
         max_icon_height = MAX(max_icon_height, sheight);
         icons_width_sum += swidth + 2;
@@ -187,22 +191,24 @@ static void node_rectangle_minimum_size(struct tree_node *node,
     
       /* buildings */
       improvement_iterate(pimprove) {
-        requirement_vector_iterate(&(pimprove->reqs), preq) {
-          if (VUT_ADVANCE == preq->source.kind
-	   && advance_number(preq->source.value.advance) == node->tech) {
-	    sprite = get_building_sprite(tileset, pimprove);
-            /* Improvement icons are not guaranteed to exist */
-            if (sprite) {
-              get_sprite_dimensions(sprite, &swidth, &sheight);
-              max_icon_height = MAX(max_icon_height, sheight);
-              icons_width_sum += swidth + 2;
+        if (valid_improvement(pimprove)) {
+          requirement_vector_iterate(&(pimprove->reqs), preq) {
+            if (VUT_ADVANCE == preq->source.kind
+                && advance_number(preq->source.value.advance) == node->tech) {
+              sprite = get_building_sprite(tileset, pimprove);
+              /* Improvement icons are not guaranteed to exist */
+              if (sprite) {
+                get_sprite_dimensions(sprite, &swidth, &sheight);
+                max_icon_height = MAX(max_icon_height, sheight);
+                icons_width_sum += swidth + 2;
+              }
             }
-	  }
-        } requirement_vector_iterate_end;
+          } requirement_vector_iterate_end;
+        }
       } improvement_iterate_end;
     
       /* governments */
-      government_iterate(gov) {
+      governments_iterate(gov) {
         requirement_vector_iterate(&(gov->reqs), preq) {
           if (VUT_ADVANCE == preq->source.kind
 	   && advance_number(preq->source.value.advance) == node->tech) {
@@ -212,7 +218,7 @@ static void node_rectangle_minimum_size(struct tree_node *node,
             icons_width_sum += swidth + 2;	    
 	  }
         } requirement_vector_iterate_end;
-      } government_iterate_end;
+      } governments_iterate_end;
     }
     
     *height += max_icon_height;
@@ -222,10 +228,10 @@ static void node_rectangle_minimum_size(struct tree_node *node,
   }
 }
 
-/****************************************************************************
+/*********************************************************************//**
   Move nodes up and down without changing order but making it more 
   symetrical. Gravitate towards parents average position.
-****************************************************************************/
+*************************************************************************/
 static void symmetrize(struct reqtree* tree)
 {
   int layer;
@@ -279,10 +285,10 @@ static void symmetrize(struct reqtree* tree)
   }
 }
 
-/****************************************************************************
+/*********************************************************************//**
   Calculate rectangles position and size from the tree.
   Logical order should already be calculated.
-****************************************************************************/
+*************************************************************************/
 static void calculate_diagram_layout(struct reqtree *tree)
 {
   int i, layer, layer_offs;
@@ -292,12 +298,12 @@ static void calculate_diagram_layout(struct reqtree *tree)
     struct tree_node *node = tree->nodes[i];
 
     node_rectangle_minimum_size(tree->nodes[i],
-				&node->node_width, &node->node_height);
+                                &node->node_width, &node->node_height);
     node->number = i;
   }
-  
+
   /* calculate height of the diagram. There should be at least 10 pixels
-   * beetween any two nodes */
+   * between any two nodes */
   tree->diagram_height = 0;
   for (layer = 0; layer < tree->num_layers; layer++) {
     int h_sum = 0;
@@ -312,7 +318,7 @@ static void calculate_diagram_layout(struct reqtree *tree)
     }
     tree->diagram_height = MAX(tree->diagram_height, h_sum);
   }
-  
+
   /* calculate maximum width of node for each layer and enlarge other nodes
    * to match maximum width
    * calculate x offsets
@@ -326,7 +332,7 @@ static void calculate_diagram_layout(struct reqtree *tree)
 
       max_width = MAX(max_width, node->node_width);
     }
-    
+
     for (i = 0; i < tree->layer_size[layer]; i++) {
       struct tree_node *node = tree->layers[layer][i];
 
@@ -374,39 +380,43 @@ static void calculate_diagram_layout(struct reqtree *tree)
   }
 }
 
-
-/*************************************************************************
-  Create a "dummy" tech tree from current ruleset.  This tree is then
-  fleshed out further (see create_reqtree). This tree doesn't include
+/*********************************************************************//**
+  Create a "dummy" tech tree from current ruleset. This tree is then
+  fleshed out further (see create_reqtree() ). This tree doesn't include
   dummy edges. Layering and ordering isn't done also.
 
   If pplayer is given, add only techs reachable by that player to tree.
 *************************************************************************/
-static struct reqtree *create_dummy_reqtree(struct player *pplayer)
+static struct reqtree *create_dummy_reqtree(struct player *pplayer,
+                                            bool show_all)
 {
+  const struct research *presearch = research_get(pplayer);
   struct reqtree *tree = fc_malloc(sizeof(*tree));
   int j;
-  struct tree_node *nodes[advance_count()];
+  Tech_type_id ac = advance_count();
+  struct tree_node *nodes[ac];
 
   nodes[A_NONE] = NULL;
-  advance_index_iterate(A_FIRST, tech) {
+  advance_index_iterate_max(A_FIRST, tech, ac) {
     if (!valid_advance_by_number(tech)) {
       nodes[tech] = NULL;
       continue;
     }
-    if (pplayer && !player_invention_reachable(pplayer, tech)) {
+    if (pplayer && !show_all
+        && !research_invention_reachable(presearch, tech)) {
       /* Reqtree requested for particular player and this tech is
-       * unreachable to him/her. */
+       * unreachable to them. */
       nodes[tech] = NULL;
       continue;
     }
     nodes[tech] = new_tree_node();
     nodes[tech]->is_dummy = FALSE;
     nodes[tech]->tech = tech;
-  } advance_index_iterate_end;
+  } advance_index_iterate_max_end;
 
-  advance_index_iterate(A_FIRST, tech) {
+  advance_index_iterate_max(A_FIRST, tech, ac) {
     struct advance *padvance = valid_advance_by_number(tech);
+    Tech_type_id tech_one, tech_two;
 
     if (!padvance) {
       continue;
@@ -415,37 +425,46 @@ static struct reqtree *create_dummy_reqtree(struct player *pplayer)
       continue;
     }
 
-    /* Formerly, we used to remove the redundant requirement nodes (the
-     * technologies already included in the requirements of the other
-     * requirement).  However, it doesn't look like a good idea, because
+    tech_one = advance_required(tech, AR_ONE);
+    tech_two = advance_required(tech, AR_TWO);
+
+    if (!show_all && A_NONE != tech_one
+        && A_LAST != tech_two && A_NONE != tech_two
+        && (nodes[tech_one] == NULL || nodes[tech_two] == NULL)) {
+      /* Print only reachable techs. */
+      continue;
+    }
+
+    /* Formerly, we used to remove the redundant requirement nodes
+     * (the technologies already included in the requirements of the other
+     * requirement). However, it doesn't look like a good idea, because
      * a player can steal any technology independently of the technology
      * tree. */
-    if (A_NONE != advance_required(tech, AR_ONE)
-        && A_LAST != advance_required(tech, AR_TWO)) {
-      add_requirement(nodes[tech], nodes[advance_required(tech, AR_ONE)]);
-      if (A_NONE != advance_required(tech, AR_TWO)) {
-        add_requirement(nodes[tech], nodes[advance_required(tech, AR_TWO)]);
+    if (A_NONE != tech_one && A_LAST != tech_two) {
+      add_requirement(nodes[tech], nodes[tech_one]);
+      if (A_NONE != tech_two) {
+        add_requirement(nodes[tech], nodes[tech_two]);
       }
     }
-  } advance_index_iterate_end;
+  } advance_index_iterate_max_end;
 
-  /* Copy nodes from local array to dynamically allocated one. 
+  /* Copy nodes from local array to dynamically allocated one.
    * Skip non-existing entries */
-  tree->nodes = fc_calloc(advance_count(), sizeof(*tree->nodes));
+  tree->nodes = fc_calloc(ac, sizeof(*tree->nodes));
   j = 0;
-  advance_index_iterate(A_FIRST, tech) {
+  advance_index_iterate_max(A_FIRST, tech, ac) {
     if (nodes[tech]) {
-      assert(valid_advance_by_number(nodes[tech]->tech));
+      fc_assert_action(valid_advance_by_number(nodes[tech]->tech), continue);
       tree->nodes[j++] = nodes[tech];
     }
-  } advance_index_iterate_end;
+  } advance_index_iterate_max_end;
   tree->num_nodes = j;
   tree->layers = NULL;
 
   return tree;
 }
 
-/*************************************************************************
+/*********************************************************************//**
   Free all memory used by tech_tree struct
 *************************************************************************/
 void destroy_reqtree(struct reqtree *tree)
@@ -469,7 +488,7 @@ void destroy_reqtree(struct reqtree *tree)
   free(tree);
 }
 
-/*************************************************************************
+/*********************************************************************//**
   Compute the longest path from this tree_node to the node with 
   no requirements. Store the result in node->layer.
 *************************************************************************/
@@ -488,7 +507,7 @@ static int longest_path(struct tree_node *node)
   return node->layer;
 }
 
-/*************************************************************************
+/*********************************************************************//**
   Compute longest_path for all nodes, thus prepare longest path layering
 *************************************************************************/
 static void longest_path_layering(struct reqtree *tree)
@@ -502,7 +521,7 @@ static void longest_path_layering(struct reqtree *tree)
   }
 }
 
-/*************************************************************************
+/*********************************************************************//**
   Find the largest value of layer amongst children of the given node
 *************************************************************************/
 static int max_provide_layer(struct tree_node *node)
@@ -518,7 +537,7 @@ static int max_provide_layer(struct tree_node *node)
   return max;
 }
 
-/*************************************************************************
+/*********************************************************************//**
   Create new tree which has dummy nodes added. The source tree is 
   completely copied, you can freely deallocate it.
 *************************************************************************/
@@ -569,7 +588,7 @@ static struct reqtree *add_dummy_nodes(struct reqtree *tree)
     struct tree_node *node = tree->nodes[i];
     int mpl;
 
-    assert(!node->is_dummy);
+    fc_assert_action(!node->is_dummy, continue);
 
     mpl = max_provide_layer(node);
 
@@ -602,7 +621,7 @@ static struct reqtree *add_dummy_nodes(struct reqtree *tree)
 
     if (mpl > node->layer + 1) {
       k += mpl - node->layer - 1;
-      assert(k <= new_tree->num_nodes);
+      fc_assert(k <= new_tree->num_nodes);
     }
   }
   new_tree->layers = NULL;
@@ -610,7 +629,7 @@ static struct reqtree *add_dummy_nodes(struct reqtree *tree)
   return new_tree;
 }
 
-/*************************************************************************
+/*********************************************************************//**
   Calculate layers[] and layer_size[] fields of tree.
   There should be layer value calculated for each node.
   Nodes will be put into layers in no particular order.
@@ -620,7 +639,7 @@ static void set_layers(struct reqtree *tree)
   int i;
   int num_layers = 0;
   
-  /* count total number of layers */
+  /* Count total number of layers */
   for (i = 0; i < tree->num_nodes; i++) {
     num_layers = MAX(num_layers, tree->nodes[i]->layer);
   }
@@ -642,8 +661,8 @@ static void set_layers(struct reqtree *tree)
     }
 
     for (i = 0; i < num_layers; i++) {
-      tree->layers[i] =
-	  fc_malloc(sizeof(*tree->layers[i]) * tree->layer_size[i]);
+      tree->layers[i]
+        = fc_malloc(sizeof(*tree->layers[i]) * tree->layer_size[i]);
     }
     for (i = 0; i < tree->num_nodes; i++) {
       struct tree_node *node = tree->nodes[i];
@@ -660,7 +679,7 @@ struct node_and_float {
   float value;
 };
 
-/*************************************************************************
+/*********************************************************************//**
   Comparison function used by barycentric_sort.
 *************************************************************************/
 static int cmp_func(const void *_a, const void *_b)
@@ -676,40 +695,42 @@ static int cmp_func(const void *_a, const void *_b)
   return 0;
 }
 
-/*************************************************************************
+/*********************************************************************//**
   Simple heuristic: Sort nodes on the given layer by the average x-value
-  of its' parents.
+  of its parents.
 *************************************************************************/
 static void barycentric_sort(struct reqtree *tree, int layer)
 {
-  struct node_and_float T[tree->layer_size[layer]];
-  int i, j;
-  float v;
+  if (tree->layer_size[layer] > 0) {
+    struct node_and_float T[tree->layer_size[layer]];
+    int i, j;
+    float v;
 
-  for (i = 0; i < tree->layer_size[layer]; i++) {
-    struct tree_node *node = tree->layers[layer][i];
+    for (i = 0; i < tree->layer_size[layer]; i++) {
+      struct tree_node *node = tree->layers[layer][i];
 
-    T[i].node = node;
-    v = 0.0;
-    for (j = 0; j < node->nrequire; j++) {
-      v += node->require[j]->order;
+      T[i].node = node;
+      v = 0.0;
+      for (j = 0; j < node->nrequire; j++) {
+        v += node->require[j]->order;
+      }
+      if (node->nrequire > 0) {
+        v /= (float) node->nrequire;
+      }
+      T[i].value = v;
     }
-    if (node->nrequire > 0) {
-      v /= (float) node->nrequire;
-    }
-    T[i].value = v;
-  }
-  qsort(T, tree->layer_size[layer], sizeof(*T),
-	cmp_func);
+    qsort(T, tree->layer_size[layer], sizeof(*T),
+          cmp_func);
 
-  for (i = 0; i < tree->layer_size[layer]; i++) {
-    tree->layers[layer][i] = T[i].node;
-    T[i].node->order = i;
+    for (i = 0; i < tree->layer_size[layer]; i++) {
+      tree->layers[layer][i] = T[i].node;
+      T[i].node->order = i;
+    }
   }
 }
 
-/*************************************************************************
-  Calculate number of edge crossings beetwen layer and layer+1
+/*********************************************************************//**
+  Calculate number of edge crossings between layer and layer + 1
 *************************************************************************/
 static int count_crossings(struct reqtree *tree, int layer)
 {
@@ -739,7 +760,7 @@ static int count_crossings(struct reqtree *tree, int layer)
   return sum;
 }
 
-/*************************************************************************
+/*********************************************************************//**
   Swap positions of two nodes on the same layer
 *************************************************************************/
 static void swap(struct reqtree *tree, int layer, int order1, int order2)
@@ -753,82 +774,83 @@ static void swap(struct reqtree *tree, int layer, int order1, int order2)
   node2->order = order1;
 }
 
-/*************************************************************************
+/*********************************************************************//**
   Try to reduce the number of crossings by swapping two nodes and checking
   if it improves the situation.
 *************************************************************************/
 static void improve(struct reqtree *tree)
 {
-  int crossings[tree->num_layers - 1];
+  int layers = tree->num_layers;
+  int crossings[layers - 1];
   int i, x1, x2, layer;
 
-  for (i = 0; i < tree->num_layers - 1; i++) {
+  for (i = 0; i < layers - 1; i++) {
     crossings[i] = count_crossings(tree, i);
   }
 
-  for (layer = 0; layer < tree->num_layers; layer++) {
+  for (layer = 0; layer < layers; layer++) {
     int layer_size = tree->layer_size[layer];
     int layer_sum = 0;
 
     if (layer > 0) {
       layer_sum += crossings[layer - 1];
     }
-    if (layer < tree->num_layers - 1) {
+    if (layer < layers - 1) {
       layer_sum += crossings[layer];
     }
 
     for (x1 = 0; x1 < layer_size; x1++) {
       for (x2 = x1 + 1; x2 < layer_size; x2++) {
-	int new_crossings = 0;
-	int new_crossings_before = 0;
+        int new_crossings = 0;
+        int new_crossings_before = 0;
 
-	swap(tree, layer, x1, x2);
-	if (layer > 0) {
-	  new_crossings_before += count_crossings(tree, layer - 1);
-	}
-	if (layer < tree->num_layers - 1) {
-	  new_crossings += count_crossings(tree, layer);
-	}
-	if (new_crossings + new_crossings_before > layer_sum) {
-	  swap(tree, layer, x1, x2);
-	} else {
-	  layer_sum = new_crossings + new_crossings_before;
-	  if (layer > 0) {
-	    crossings[layer - 1] = new_crossings_before;
-	  }
-	  if (layer < tree->num_layers - 1) {
-	    crossings[layer] = new_crossings;
-	  }
-	}
+        swap(tree, layer, x1, x2);
+        if (layer > 0) {
+          new_crossings_before += count_crossings(tree, layer - 1);
+        }
+        if (layer < layers - 1) {
+          new_crossings += count_crossings(tree, layer);
+        }
+        if (new_crossings + new_crossings_before > layer_sum) {
+          swap(tree, layer, x1, x2);
+        } else {
+          layer_sum = new_crossings + new_crossings_before;
+          if (layer > 0) {
+            crossings[layer - 1] = new_crossings_before;
+          }
+          if (layer < layers - 1) {
+            crossings[layer] = new_crossings;
+          }
+        }
       }
     }
   }
 }
 
-/*************************************************************************
+/*********************************************************************//**
   Generate optimized tech_tree from current ruleset.
-  You should free it by destroy_reqtree.
+  You should free it by destroy_reqtree().
 
   If pplayer is not NULL, techs unreachable to that player are not shown.
 *************************************************************************/
-struct reqtree *create_reqtree(struct player *pplayer)
+struct reqtree *create_reqtree(struct player *pplayer, bool show_all)
 {
   struct reqtree *tree1, *tree2;
   int i, j;
 
-  tree1 = create_dummy_reqtree(pplayer);
+  tree1 = create_dummy_reqtree(pplayer, show_all);
   longest_path_layering(tree1);
   tree2 = add_dummy_nodes(tree1);
   destroy_reqtree(tree1);
   set_layers(tree2);
-  
+
   /* It's good heuristics for beginning */
   for (j = 0; j < 20; j++) {
     for (i = 0; i < tree2->num_layers; i++) {
       barycentric_sort(tree2, i);
     }
   }
-  
+
   /* Now burn some CPU */
   for (j = 0; j < 20; j++) {
     improve(tree2);
@@ -839,57 +861,65 @@ struct reqtree *create_reqtree(struct player *pplayer)
   return tree2;
 }
 
-/****************************************************************************
+/*********************************************************************//**
   Give the dimensions of the reqtree.
-****************************************************************************/
+*************************************************************************/
 void get_reqtree_dimensions(struct reqtree *reqtree,
-			    int *width, int *height)
+                            int *width, int *height)
 {
   if (width) {
     *width = reqtree->diagram_width;
   }
-  if (height){
+  if (height) {
     *height = reqtree->diagram_height;
   }
 }
 
-/****************************************************************************
+/*********************************************************************//**
   Return a background color of node's rectangle
-****************************************************************************/
+*************************************************************************/
 static enum color_std node_color(struct tree_node *node)
 {
   if (!node->is_dummy) {
-    struct player_research* research = get_player_research(client.conn.playing);
+    struct research *research = research_get(client_player());
 
     if (!research) {
       return COLOR_REQTREE_KNOWN;
     }
 
-    if (!player_invention_reachable(client.conn.playing, node->tech)) {
+    if (!research_invention_reachable(research, node->tech)) {
       return COLOR_REQTREE_UNREACHABLE;
     }
 
+    if (!research_invention_gettable(research, node->tech, TRUE)) {
+      if (research_goal_tech_req(research, research->tech_goal, node->tech)
+          || node->tech == research->tech_goal) {
+        return COLOR_REQTREE_GOAL_NOT_GETTABLE;
+      } else {
+        return COLOR_REQTREE_NOT_GETTABLE;
+      }
+    }
+
     if (research->researching == node->tech) {
       return COLOR_REQTREE_RESEARCHING;
     }
-    
-    if (TECH_KNOWN == player_invention_state(client.conn.playing, node->tech)) {
+
+    if (TECH_KNOWN == research_invention_state(research, node->tech)) {
       return COLOR_REQTREE_KNOWN;
     }
 
-    if (is_tech_a_req_for_goal(client.conn.playing, node->tech,
-                               research->tech_goal)
+    if (research_goal_tech_req(research, research->tech_goal, node->tech)
 	|| node->tech == research->tech_goal) {
-      if (TECH_PREREQS_KNOWN ==
-            player_invention_state(client.conn.playing, node->tech)) {
+      if (TECH_PREREQS_KNOWN == research_invention_state(research,
+                                                         node->tech)) {
 	return COLOR_REQTREE_GOAL_PREREQS_KNOWN;
       } else {
 	return COLOR_REQTREE_GOAL_UNKNOWN;
       }
     }
 
-    if (TECH_PREREQS_KNOWN ==
-          player_invention_state(client.conn.playing, node->tech)) {
+    if (TECH_PREREQS_KNOWN == research_invention_state(research,
+                                                       node->tech)) {
       return COLOR_REQTREE_PREREQS_KNOWN;
     }
 
@@ -900,15 +930,14 @@ static enum color_std node_color(struct tree_node *node)
 
 }
 
-
-/****************************************************************************
+/*********************************************************************//**
   Return the type for an edge between two nodes
   if node is a dummy, dest_node can be NULL
-****************************************************************************/
-static enum reqtree_edge_type get_edge_type(struct tree_node *node, 
+*************************************************************************/
+static enum reqtree_edge_type get_edge_type(struct tree_node *node,
                                             struct tree_node *dest_node)
 {
-  struct player_research *research = get_player_research(client.conn.playing);
+  struct research *research = research_get(client_player());
 
   if (dest_node == NULL) {
     /* assume node is a dummy */
@@ -917,7 +946,7 @@ static enum reqtree_edge_type get_edge_type(struct tree_node *node,
    
   /* find the required tech */
   while (node->is_dummy) {
-    assert(node->nrequire == 1);
+    fc_assert(node->nrequire == 1);
     node = node->require[0];
   }
   
@@ -927,7 +956,7 @@ static enum reqtree_edge_type get_edge_type(struct tree_node *node,
     enum reqtree_edge_type sum_type = REQTREE_EDGE;
     int i;
 
-    assert(dest_node->nprovide > 0);
+    fc_assert(dest_node->nprovide > 0);
     for (i = 0; i < dest_node->nprovide; ++i) {
       enum reqtree_edge_type type = get_edge_type(node, dest_node->provide[i]);
       switch (type) {
@@ -955,14 +984,13 @@ static enum reqtree_edge_type get_edge_type(struct tree_node *node,
     return REQTREE_ACTIVE_EDGE;
   }
 
-  if (is_tech_a_req_for_goal(client.conn.playing, dest_node->tech,
-                             research->tech_goal)
+  if (research_goal_tech_req(research, research->tech_goal, dest_node->tech)
       || dest_node->tech == research->tech_goal) {
     return REQTREE_GOAL_EDGE;
   }
 
-  if (TECH_KNOWN == player_invention_state(client.conn.playing, node->tech)) {
-    if (TECH_KNOWN == player_invention_state(client.conn.playing, dest_node->tech)) {
+  if (TECH_KNOWN == research_invention_state(research, node->tech)) {
+    if (TECH_KNOWN == research_invention_state(research, dest_node->tech)) {
       return REQTREE_KNOWN_EDGE;
     } else {
       return REQTREE_READY_EDGE;
@@ -972,14 +1000,15 @@ static enum reqtree_edge_type get_edge_type(struct tree_node *node,
   return REQTREE_EDGE;
 }
 
-/****************************************************************************
+/*********************************************************************//**
   Return a stroke color for an edge between two nodes
   if node is a dummy, dest_node can be NULL
-****************************************************************************/
+*************************************************************************/
 static enum color_std edge_color(struct tree_node *node, 
                                  struct tree_node *dest_node)
 {
   enum reqtree_edge_type type = get_edge_type(node, dest_node);
+
   switch (type) {
   case REQTREE_ACTIVE_EDGE:
     return COLOR_REQTREE_RESEARCHING;
@@ -995,15 +1024,15 @@ static enum color_std edge_color(struct tree_node *node,
   };
 }
 
-/****************************************************************************
+/*********************************************************************//**
   Draw the reqtree diagram!
 
   This draws the given portion of the reqtree diagram (given by
   (tt_x,tt_y) and (w,h) onto the canvas at position (canvas_x, canvas_y).
-****************************************************************************/
+*************************************************************************/
 void draw_reqtree(struct reqtree *tree, struct canvas *pcanvas,
-		  int canvas_x, int canvas_y,
-		  int tt_x, int tt_y, int w, int h)
+                  int canvas_x, int canvas_y,
+                  int tt_x, int tt_y, int w, int h)
 {
   int i, j, k;
   int swidth, sheight;
@@ -1028,10 +1057,11 @@ void draw_reqtree(struct reqtree *tree, struct canvas *pcanvas,
 		        LINE_GOTO,
 		        startx, starty, width, 0);
       } else {
-	const char *text = advance_name_for_player(client.conn.playing, node->tech);
+        const char *text = research_advance_name_translation
+                               (research_get(client_player()), node->tech);
 	int text_w, text_h;
 	int icon_startx;
-	
+
         canvas_put_rectangle(pcanvas,
                              get_color(tileset, COLOR_REQTREE_BACKGROUND),
                              startx, starty, width, height);
@@ -1044,7 +1074,7 @@ void draw_reqtree(struct reqtree *tree, struct canvas *pcanvas,
 	 * node_rectangle_minimum_size(). If you change something here,
 	 * change also node_rectangle_minimum_size().
 	 */
-			     
+
 	get_text_size(&text_w, &text_h, FONT_REQTREE_TEXT, text);
 
 	canvas_put_text(pcanvas,
@@ -1055,12 +1085,12 @@ void draw_reqtree(struct reqtree *tree, struct canvas *pcanvas,
 			text);
  	icon_startx = startx + 5;
 	
-	if (reqtree_show_icons) {
+	if (gui_options.reqtree_show_icons) {
  	  unit_type_iterate(unit) {
             if (advance_number(unit->require_advance) != node->tech) {
 	      continue;
 	    }
- 	    sprite = get_unittype_sprite(tileset, unit);
+ 	    sprite = get_unittype_sprite(tileset, unit, direction8_invalid());
  	    get_sprite_dimensions(sprite, &swidth, &sheight);
  	    canvas_put_sprite_full(pcanvas,
  	                           icon_startx,
@@ -1071,25 +1101,27 @@ void draw_reqtree(struct reqtree *tree, struct canvas *pcanvas,
  	  } unit_type_iterate_end;
        
           improvement_iterate(pimprove) {
-            requirement_vector_iterate(&(pimprove->reqs), preq) {
-              if (VUT_ADVANCE == preq->source.kind
-	       && advance_number(preq->source.value.advance) == node->tech) {
- 	        sprite = get_building_sprite(tileset, pimprove);
-                /* Improvement icons are not guaranteed to exist */
-                if (sprite) {
-                  get_sprite_dimensions(sprite, &swidth, &sheight);
-                  canvas_put_sprite_full(pcanvas,
-                                         icon_startx,
-                                         starty + text_h + 4
-                                         + (height - text_h - 4 - sheight) / 2,
-                                         sprite);
-                  icon_startx += swidth + 2;
+            if (valid_improvement(pimprove)) {
+              requirement_vector_iterate(&(pimprove->reqs), preq) {
+                if (VUT_ADVANCE == preq->source.kind
+                    && advance_number(preq->source.value.advance) == node->tech) {
+                  sprite = get_building_sprite(tileset, pimprove);
+                  /* Improvement icons are not guaranteed to exist */
+                  if (sprite) {
+                    get_sprite_dimensions(sprite, &swidth, &sheight);
+                    canvas_put_sprite_full(pcanvas,
+                                           icon_startx,
+                                           starty + text_h + 4
+                                           + (height - text_h - 4 - sheight) / 2,
+                                           sprite);
+                    icon_startx += swidth + 2;
+                  }
                 }
- 	      }
- 	    } requirement_vector_iterate_end;
+              } requirement_vector_iterate_end;
+            }
           } improvement_iterate_end;
 
-          government_iterate(gov) {
+          governments_iterate(gov) {
             requirement_vector_iterate(&(gov->reqs), preq) {
               if (VUT_ADVANCE == preq->source.kind
                && advance_number(preq->source.value.advance) == node->tech) {
@@ -1103,7 +1135,7 @@ void draw_reqtree(struct reqtree *tree, struct canvas *pcanvas,
  	        icon_startx += swidth + 2;
               }
             } requirement_vector_iterate_end;
-          } government_iterate_end;
+          } governments_iterate_end;
         }
       }
 
@@ -1112,12 +1144,13 @@ void draw_reqtree(struct reqtree *tree, struct canvas *pcanvas,
       starty = node->node_y + node->node_height / 2;
       for (k = 0; k < node->nprovide; k++) {
 	struct tree_node *dest_node = node->provide[k];
+
 	color = get_color(tileset, edge_color(node, dest_node));
 
 	endx = dest_node->node_x;
 	endy = dest_node->node_y + dest_node->node_height / 2;
-	
-        if (reqtree_curved_lines) {
+
+        if (gui_options.reqtree_curved_lines) {
           canvas_put_curved_line(pcanvas, color, LINE_GOTO,
                                  startx, starty, endx - startx,
                                  endy - starty);
@@ -1131,9 +1164,9 @@ void draw_reqtree(struct reqtree *tree, struct canvas *pcanvas,
   }
 }
 
-/****************************************************************************
+/*********************************************************************//**
   Return the tech ID at the given position of the reqtree (or A_NONE).
-****************************************************************************/
+*************************************************************************/
 Tech_type_id get_tech_on_reqtree(struct reqtree *tree, int x, int y)
 {
   int i;
@@ -1153,12 +1186,12 @@ Tech_type_id get_tech_on_reqtree(struct reqtree *tree, int x, int y)
   return A_NONE;
 }
 
-/****************************************************************************
+/*********************************************************************//**
   Return the position of the given tech on the reqtree.  Return TRUE iff
   it was found.
-****************************************************************************/
+*************************************************************************/
 bool find_tech_on_reqtree(struct reqtree *tree, Tech_type_id tech,
-			  int *x, int *y, int *w, int *h)
+                          int *x, int *y, int *w, int *h)
 {
   int i;
 
