@@ -1,4 +1,4 @@
-/********************************************************************** 
+/***********************************************************************
  Freeciv - Copyright (C) 1996 - A Kjeldberg, L Gregersen, P Unold
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -11,7 +11,7 @@
    GNU General Public License for more details.
 ***********************************************************************/
 
-/********************************************************************** 
+/***********************************************************************
   A low-level object for reading a registry-format file.
   original author: David Pfitzner <dwp@mso.anu.edu.au>
 
@@ -25,12 +25,12 @@
   pointing to some data if the token was found, or NULL otherwise.
   The data pointed to should not be modified.  The retuned pointer
   is valid _only_ until another inputfile is performed.  (So should
-  be used immediately, or mystrdup-ed etc.)
+  be used immediately, or fc_strdup-ed etc.)
   
   The tokens recognised are as follows:
   (Single quotes are delimiters used here, but are not part of the
   actual tokens/strings.)
-  Most tokens can be preceeded by optional whitespace; exceptions
+  Most tokens can be preceded by optional whitespace; exceptions
   are section_name and entry_name.
 
   section_name:  '[foo]'
@@ -64,14 +64,16 @@
 ***********************************************************************/
 
 #ifdef HAVE_CONFIG_H
-#include <config.h>
+#include <fc_config.h>
 #endif
 
-#include <assert.h>
+#include <stdarg.h>
 #include <stdio.h>
 #include <string.h>
 
+/* utility */
 #include "astring.h"
+#include "fcintl.h"
 #include "ioz.h"
 #include "log.h"
 #include "mem.h"
@@ -90,12 +92,9 @@ struct inputfile {
   char *filename;		/* filename as passed to fopen */
   fz_FILE *fp;			/* read from this */
   bool at_eof;			/* flag for end-of-file */
-  struct astring cur_line;	/* data from current line, or .n==0 if
-				   have not yet read in the current line */
-  struct astring copy_line;	/* original cur_line (sometimes insert nulls
-				   in cur_line for processing) */
-  int cur_line_pos;		/* position in current line */
-  int line_num;			/* line number from file in cur_line */
+  struct astring cur_line;	/* data from current line */
+  unsigned int cur_line_pos;    /* position in current line */
+  unsigned int line_num;        /* line number from file in cur_line */
   struct astring token;		/* data returned to user */
   struct astring partial;	/* used in accumulating multi-line strings;
 				   used only in get_token_value, but put
@@ -140,9 +139,16 @@ tok_tab[INF_TOK_LAST] =
 };
 
 static bool read_a_line(struct inputfile *inf);
-static void inf_warn(struct inputfile *inf, const char *message);
 
-/********************************************************************** 
+#define inf_log(inf, level, message, ...)                                   \
+  if (log_do_output_for_level(level)) {                                     \
+    do_log(__FILE__, __FUNCTION__, __FC_LINE__, FALSE, level, "%s",         \
+           inf_log_str(inf, message, ## __VA_ARGS__));                      \
+  }
+#define inf_warn(inf, message)                                              \
+  inf_log(inf, LOG_NORMAL, "%s", message);
+
+/*******************************************************************//**
   Return true if c is a 'comment' character: '#' or ';'
 ***********************************************************************/
 static bool is_comment(int c)
@@ -150,13 +156,13 @@ static bool is_comment(int c)
   return (c == '#' || c == ';');
 }
 
-/********************************************************************** 
+/*******************************************************************//**
   Set values to zeros; should have free'd/closed everything before
   this if appropriate.
 ***********************************************************************/
 static void init_zeros(struct inputfile *inf)
 {
-  assert(inf != NULL);
+  fc_assert_ret(NULL != inf);
   inf->magic = INF_MAGIC;
   inf->filename = NULL;
   inf->fp = NULL;
@@ -166,43 +172,37 @@ static void init_zeros(struct inputfile *inf)
   inf->at_eof = inf->in_string = FALSE;
   inf->string_start_line = 0;
   astr_init(&inf->cur_line);
-  astr_init(&inf->copy_line);
   astr_init(&inf->token);
   astr_init(&inf->partial);
 }
 
-/********************************************************************** 
+/*******************************************************************//**
   Check sensible values for an opened inputfile.
 ***********************************************************************/
-static void assert_sanity(struct inputfile *inf)
+static bool inf_sanity_check(struct inputfile *inf)
 {
-  assert(inf != NULL);
-  assert(inf->magic==INF_MAGIC);
-  assert(inf->fp != NULL);
-  assert(inf->line_num >= 0);
-  assert(inf->cur_line_pos >= 0);
-  assert(inf->at_eof == FALSE || inf->at_eof == TRUE);
-  assert(inf->in_string == FALSE || inf->in_string == TRUE);
-#ifdef DEBUG
-  assert(inf->string_start_line >= 0);
-  assert(inf->cur_line.n >= 0);
-  assert(inf->copy_line.n >= 0);
-  assert(inf->token.n >= 0);
-  assert(inf->partial.n >= 0);
-  assert(inf->cur_line.n_alloc >= 0);
-  assert(inf->copy_line.n_alloc >= 0);
-  assert(inf->token.n_alloc >= 0);
-  assert(inf->partial.n_alloc >= 0);
-  if(inf->included_from) {
-    assert_sanity(inf->included_from);
+  fc_assert_ret_val(NULL != inf, FALSE);
+  fc_assert_ret_val(INF_MAGIC == inf->magic, FALSE);
+  fc_assert_ret_val(NULL != inf->fp, FALSE);
+  fc_assert_ret_val(!inf->at_eof
+                    || inf->at_eof, FALSE);
+  fc_assert_ret_val(!inf->in_string
+                    || inf->in_string, FALSE);
+
+#ifdef FREECIV_DEBUG
+  fc_assert_ret_val(0 <= inf->string_start_line, FALSE);
+  if (inf->included_from && !inf_sanity_check(inf->included_from)) {
+    return FALSE;
   }
-#endif
+#endif /* FREECIV_DEBUG */
+
+  return TRUE;
 }
 
-/**************************************************************************
+/*******************************************************************//**
   Return the filename the inputfile was loaded as, or "(anonymous)"
   if this inputfile was loaded from a stream rather than from a file.
-**************************************************************************/
+***********************************************************************/
 static const char *inf_filename(struct inputfile *inf)
 {
   if (inf->filename) {
@@ -212,50 +212,50 @@ static const char *inf_filename(struct inputfile *inf)
   }
 }
 
-/********************************************************************** 
+/*******************************************************************//**
   Open the file, and return an allocated, initialized structure.
   Returns NULL if the file could not be opened.
 ***********************************************************************/
 struct inputfile *inf_from_file(const char *filename,
-				datafilename_fn_t datafn)
+                                datafilename_fn_t datafn)
 {
   struct inputfile *inf;
   fz_FILE *fp;
 
-  assert(filename != NULL);
-  assert(strlen(filename) > 0);
-  fp = fz_from_file(filename, "r", FZ_NOT_USED, FZ_NOT_USED);
+  fc_assert_ret_val(NULL != filename, NULL);
+  fc_assert_ret_val(0 < strlen(filename), NULL);
+  fp = fz_from_file(filename, "r", -1, 0);
   if (!fp) {
     return NULL;
   }
-  freelog(LOG_DEBUG, "inputfile: opened \"%s\" ok", filename);
+  log_debug("inputfile: opened \"%s\" ok", filename);
   inf = inf_from_stream(fp, datafn);
-  inf->filename = mystrdup(filename);
+  inf->filename = fc_strdup(filename);
   return inf;
 }
 
-/********************************************************************** 
+/*******************************************************************//**
   Open the stream, and return an allocated, initialized structure.
   Returns NULL if the file could not be opened.
 ***********************************************************************/
-struct inputfile *inf_from_stream(fz_FILE * stream, datafilename_fn_t datafn)
+struct inputfile *inf_from_stream(fz_FILE *stream, datafilename_fn_t datafn)
 {
   struct inputfile *inf;
 
-  assert(stream != NULL);
+  fc_assert_ret_val(NULL != stream, NULL);
   inf = fc_malloc(sizeof(*inf));
   init_zeros(inf);
-  
+
   inf->filename = NULL;
   inf->fp = stream;
   inf->datafn = datafn;
 
-  freelog(LOG_DEBUG, "inputfile: opened \"%s\" ok", inf_filename(inf));
+  log_debug("inputfile: opened \"%s\" ok", inf_filename(inf));
   return inf;
 }
 
 
-/********************************************************************** 
+/*******************************************************************//**
   Close the file and free associated memory, but don't recurse
   included_from files, and don't free the actual memory where
   the inf record is stored (ie, the memory where the users pointer
@@ -263,36 +263,35 @@ struct inputfile *inf_from_stream(fz_FILE * stream, datafilename_fn_t datafn)
 ***********************************************************************/
 static void inf_close_partial(struct inputfile *inf)
 {
-  assert_sanity(inf);
+  inf_sanity_check(inf);
 
-  freelog(LOG_DEBUG, "inputfile: sub-closing \"%s\"", inf_filename(inf));
+  log_debug("inputfile: sub-closing \"%s\"", inf_filename(inf));
 
   if (fz_ferror(inf->fp) != 0) {
-    freelog(LOG_ERROR, "Error before closing %s: %s", inf_filename(inf),
-	    fz_strerror(inf->fp));
+    log_error("Error before closing %s: %s", inf_filename(inf),
+              fz_strerror(inf->fp));
     fz_fclose(inf->fp);
     inf->fp = NULL;
   }
   else if (fz_fclose(inf->fp) != 0) {
-    freelog(LOG_ERROR, "Error closing %s", inf_filename(inf));
+    log_error("Error closing %s", inf_filename(inf));
   }
   if (inf->filename) {
     free(inf->filename);
   }
   inf->filename = NULL;
   astr_free(&inf->cur_line);
-  astr_free(&inf->copy_line);
   astr_free(&inf->token);
   astr_free(&inf->partial);
 
-  /* assign zeros for safety if accidently re-use etc: */
+  /* Assign zeros for safety if accidentally reuse etc: */
   init_zeros(inf);
   inf->magic = ~INF_MAGIC;
 
-  freelog(LOG_DEBUG, "inputfile: sub-closed ok");
+  log_debug("inputfile: sub-closed ok");
 }
 
-/********************************************************************** 
+/*******************************************************************//**
   Close the file and free associated memory, included any partially
   recursed included files, and the memory allocated for 'inf' itself.
   Should only be used on an actually open inputfile.
@@ -300,46 +299,60 @@ static void inf_close_partial(struct inputfile *inf)
 ***********************************************************************/
 void inf_close(struct inputfile *inf)
 {
-  assert_sanity(inf);
+  inf_sanity_check(inf);
+
+  log_debug("inputfile: closing \"%s\"", inf_filename(inf));
 
-  freelog(LOG_DEBUG, "inputfile: closing \"%s\"", inf_filename(inf));
   if (inf->included_from) {
     inf_close(inf->included_from);
+    /* Stop anything from recursing to already closed including file */
+    inf->included_from = NULL;
   }
   inf_close_partial(inf);
   free(inf);
-  freelog(LOG_DEBUG, "inputfile: closed ok");
+
+  log_debug("inputfile: closed ok");
 }
 
-/********************************************************************** 
-  Return 1 if have data for current line.
+/*******************************************************************//**
+  Return TRUE if have data for current line.
 ***********************************************************************/
 static bool have_line(struct inputfile *inf)
 {
-  assert_sanity(inf);
-  return (inf->cur_line.n > 0);
+  if (!inf_sanity_check(inf)) {
+    return FALSE;
+  }
+
+  return !astr_empty(&inf->cur_line);
 }
 
-/********************************************************************** 
-  Return 1 if current pos is at end of current line.
+/*******************************************************************//**
+  Return TRUE if current pos is at end of current line.
 ***********************************************************************/
 static bool at_eol(struct inputfile *inf)
 {
-  assert_sanity(inf);
-  assert(inf->cur_line_pos <= inf->cur_line.n);
-  return (inf->cur_line_pos >= inf->cur_line.n - 1);
+  if (!inf_sanity_check(inf)) {
+    return TRUE;
+  }
+
+  fc_assert_ret_val(inf->cur_line_pos <= astr_len(&inf->cur_line), TRUE);
+
+  return (inf->cur_line_pos >= astr_len(&inf->cur_line));
 }
 
-/********************************************************************** 
-  ...
+/*******************************************************************//**
+  Return TRUE if current pos is at end of file.
 ***********************************************************************/
 bool inf_at_eof(struct inputfile *inf)
 {
-  assert_sanity(inf);
+  if (!inf_sanity_check(inf)) {
+    return TRUE;
+  }
+
   return inf->at_eof;
 }
 
-/********************************************************************** 
+/*******************************************************************//**
   Check for an include command, which is an isolated line with:
      *include "filename"
   If a file is included via this mechanism, returns 1, and sets up
@@ -351,72 +364,90 @@ static bool check_include(struct inputfile *inf)
 {
   const char *include_prefix = "*include";
   static size_t len = 0;
-  char *bare_name, *full_name, *c;
+  size_t bare_name_len;
+  char *bare_name;
+  const char *c, *bare_name_start, *full_name;
   struct inputfile *new_inf, temp;
 
-  if (len==0) {
+  if (len == 0) {
     len = strlen(include_prefix);
   }
-  assert_sanity(inf);
-  if (inf->in_string || inf->cur_line.n <= len || inf->cur_line_pos > 0) {
+
+  if (!inf_sanity_check(inf)) {
     return FALSE;
   }
-  if (strncmp(inf->cur_line.str, include_prefix, len)!=0) {
+
+  if (inf->in_string || astr_len(&inf->cur_line) <= len
+      || inf->cur_line_pos > 0) {
     return FALSE;
   }
-  /* from here, the include-line must be well formed or we die */
+
+  if (strncmp(astr_str(&inf->cur_line), include_prefix, len) != 0) {
+    return FALSE;
+  }
+
+  /* from here, the include-line must be well formed */
   /* keep inf->cur_line_pos accurate just so error messages are useful */
 
   /* skip any whitespace: */
   inf->cur_line_pos = len;
-  c = inf->cur_line.str + len;
-  while (*c != '\0' && my_isspace(*c)) c++;
+  c = astr_str(&inf->cur_line) + len;
+  while (*c != '\0' && fc_isspace(*c)) {
+    c++;
+  }
 
   if (*c != '\"') {
-    inf_log(inf, LOG_ERROR, 
+    inf_log(inf, LOG_ERROR,
             "Did not find opening doublequote for '*include' line");
     return FALSE;
   }
   c++;
-  inf->cur_line_pos = c - inf->cur_line.str;
+  inf->cur_line_pos = c - astr_str(&inf->cur_line);
 
-  bare_name = c;
+  bare_name_start = c;
   while (*c != '\0' && *c != '\"') c++;
   if (*c != '\"') {
     inf_log(inf, LOG_ERROR, 
             "Did not find closing doublequote for '*include' line");
     return FALSE;
   }
-  *c++ = '\0';
-  inf->cur_line_pos = c - inf->cur_line.str;
+  c++;
+  bare_name_len = c - bare_name_start;
+  bare_name = fc_malloc(bare_name_len);
+  strncpy(bare_name, bare_name_start, bare_name_len - 1);
+  bare_name[bare_name_len - 1] = '\0';
+  inf->cur_line_pos = c - astr_str(&inf->cur_line);
 
   /* check rest of line is well-formed: */
-  while (*c != '\0' && my_isspace(*c) && !is_comment(*c)) c++;
-  if (!(*c=='\0' || is_comment(*c))) {
+  while (*c != '\0' && fc_isspace(*c) && !is_comment(*c)) {
+    c++;
+  }
+  if (!(*c == '\0' || is_comment(*c))) {
     inf_log(inf, LOG_ERROR, "Junk after filename for '*include' line");
     return FALSE;
   }
-  inf->cur_line_pos = inf->cur_line.n-1;
+  inf->cur_line_pos = astr_len(&inf->cur_line) - 1;
 
   full_name = inf->datafn(bare_name);
   if (!full_name) {
-    freelog(LOG_ERROR, "Could not find included file \"%s\"", bare_name);
+    log_error("Could not find included file \"%s\"", bare_name);
+    free(bare_name);
     return FALSE;
   }
+  free(bare_name);
 
   /* avoid recursion: (first filename may not have the same path,
-     but will at least stop infinite recursion) */
+   * but will at least stop infinite recursion) */
   {
     struct inputfile *inc = inf;
     do {
-      if (inc->filename && strcmp(full_name, inc->filename)==0) {
-        freelog(LOG_ERROR, 
-                "Recursion trap on '*include' for \"%s\"", full_name);
+      if (inc->filename && strcmp(full_name, inc->filename) == 0) {
+        log_error("Recursion trap on '*include' for \"%s\"", full_name);
         return FALSE;
       }
-    } while((inc=inc->included_from));
+    } while ((inc = inc->included_from));
   }
-  
+
   new_inf = inf_from_file(full_name, inf->datafn);
 
   /* Swap things around so that memory pointed to by inf (user pointer,
@@ -431,10 +462,10 @@ static bool check_include(struct inputfile *inf)
   return TRUE;
 }
 
-/********************************************************************** 
-  Read a new line into cur_line; also copy to copy_line.
+/*******************************************************************//**
+  Read a new line into cur_line.
   Increments line_num and cur_line_pos.
-  Returns 0 if didn't read or other problem: treat as EOF.
+  Returns FALSE if didn't read or other problem: treat as EOF.
   Strips newline from input.
 ***********************************************************************/
 static bool read_a_line(struct inputfile *inf)
@@ -442,31 +473,36 @@ static bool read_a_line(struct inputfile *inf)
   struct astring *line;
   char *ret;
   int pos;
-  
-  assert_sanity(inf);
 
-  if (inf->at_eof)
+  if (!inf_sanity_check(inf)) {
     return FALSE;
-  
+  }
+
+  if (inf->at_eof) {
+    return FALSE;
+  }
+
   /* abbreviation: */
   line = &inf->cur_line;
-  
+
   /* minimum initial line length: */
-  astr_minsize(line, 80);
+  astr_reserve(line, 80);
+  astr_clear(line);
   pos = 0;
 
-  /* don't print "orig line" in warnings until we have it: */
-  inf->copy_line.n = 0;
-  
   /* Read until we get a full line:
    * At start of this loop, pos is index to trailing null
    * (or first position) in line.
    */
-  for(;;) {
-    ret = fz_fgets(line->str + pos, line->n_alloc - pos, inf->fp);
-    
+  for (;;) {
+    ret = fz_fgets((char *) astr_str(line) + pos,
+                   astr_capacity(line) - pos, inf->fp);
+
     if (!ret) {
       /* fgets failed */
+      if (pos > 0) {
+        inf_log(inf, LOG_ERROR, _("End-of-file not in line of its own"));
+      }
       inf->at_eof = TRUE;
       if (inf->in_string) {
         /* Note: Don't allow multi-line strings to cross "include"
@@ -476,33 +512,40 @@ static bool read_a_line(struct inputfile *inf)
       }
       break;
     }
-    
-    pos += strlen(line->str + pos);
-    line->n = pos + 1;
-    
-    if (line->str[pos-1] == '\n') {
-      line->str[pos-1] = '\0';
-      line->n--;
+
+    /* Cope with \n\r line endings if not caught by library:
+     * strip off any leading \r */
+    if (0 == pos && 0 < astr_len(line) && astr_str(line)[0] == '\r') {
+      memmove((char *)astr_str(line), astr_str(line)+1, astr_len(line));
+    }
+
+    pos = astr_len(line);
+
+    if (0 < pos && astr_str(line)[pos - 1] == '\n') {
+      int end;
+      /* Cope with \r\n line endings if not caught by library:
+       * strip off any trailing \r */
+      if (1 < pos && astr_str(line)[pos - 2] == '\r') {
+        end = pos - 2;
+      } else {
+        end = pos - 1;
+      }
+      *((char *) astr_str(line) + end) = '\0';
       break;
     }
-    if (line->n != line->n_alloc) {
-      freelog(LOG_VERBOSE, "inputfile: expect missing newline at EOF");
+    astr_reserve(line, pos * 2);
+  }
+
+  if (!inf->at_eof) {
+    inf->line_num++;
+    inf->cur_line_pos = 0;
+
+    if (check_include(inf)) {
+      return read_a_line(inf);
     }
-    astr_minsize(line, line->n*2);
-  }
-  inf->line_num++;
-  inf->cur_line_pos = 0;
-
-  astr_minsize(&inf->copy_line, inf->cur_line.n + ((inf->cur_line.n == 0) ? 1 : 0));
-  strcpy(inf->copy_line.str, inf->cur_line.str);
-
-  if (check_include(inf)) {
-    return read_a_line(inf);
-  }
-
-  if (inf->at_eof) {
-    line->str[0] = '\0';
-    line->n = 0;
+    return TRUE;
+  } else {
+    astr_clear(line);
     if (inf->included_from) {
       /* Pop the include, and get next line from file above instead. */
       struct inputfile *inc = inf->included_from;
@@ -513,87 +556,76 @@ static bool read_a_line(struct inputfile *inf)
       return read_a_line(inf);
     }
     return FALSE;
-  } else {
-    return TRUE;
   }
 }
 
-/********************************************************************** 
-  Set "flag" token when we don't really want to return anything,
-  except non-null.
+/*******************************************************************//**
+  Return a detailed log message, including information on current line
+  number etc. Message can be NULL: then just logs information on where
+  we are in the file.
 ***********************************************************************/
-static void assign_flag_token(struct astring *astr, char val)
+char *inf_log_str(struct inputfile *inf, const char *message, ...)
 {
-  static char flag_token[2];
-
-  assert(astr != NULL);
-  flag_token[0] = val;
-  flag_token[1] = '\0';
-  astr_minsize(astr, 2);
-  strcpy(astr->str, flag_token);
-}
-
-/********************************************************************** 
-  Give a detailed log message, including information on
-  current line number etc.  Message can be NULL: then just logs
-  information on where we are in the file.
-***********************************************************************/
-void inf_log(struct inputfile *inf, int loglevel, const char *message)
-{
-  assert_sanity(inf);
+  va_list args;
+  static char str[512];
 
   if (message) {
-    freelog(loglevel, "%s", message);
+    va_start(args, message);
+    fc_vsnprintf(str, sizeof(str), message, args);
+    va_end(args);
+    sz_strlcat(str, "\n");
+  } else {
+    str[0] = '\0';
   }
-  freelog(loglevel, "  file \"%s\", line %d, pos %d%s",
-	  inf_filename(inf), inf->line_num, inf->cur_line_pos,
-	  (inf->at_eof ? ", EOF" : ""));
-  if (inf->cur_line.str && inf->cur_line.n > 0) {
-    freelog(loglevel, "  looking at: '%s'",
-	    inf->cur_line.str+inf->cur_line_pos);
-  }
-  if (inf->copy_line.str && inf->copy_line.n > 0) {
-    freelog(loglevel, "  original line: '%s'", inf->copy_line.str);
-  }
-  if (inf->in_string) {
-    freelog(loglevel, "  processing string starting at line %d",
-	    inf->string_start_line);
-  }
-  while ((inf=inf->included_from)) {    /* local pointer assignment */
-    freelog(loglevel, "  included from file \"%s\", line %d",
-	    inf_filename(inf), inf->line_num);
+
+  if (inf_sanity_check(inf)) {
+    cat_snprintf(str, sizeof(str), "  file \"%s\", line %d, pos %d%s",
+                 inf_filename(inf), inf->line_num, inf->cur_line_pos,
+                 (inf->at_eof ? ", EOF" : ""));
+
+    if (!astr_empty(&inf->cur_line)) {
+      cat_snprintf(str, sizeof(str), "\n  looking at: '%s'",
+                   astr_str(&inf->cur_line) + inf->cur_line_pos);
+    }
+    if (inf->in_string) {
+      cat_snprintf(str, sizeof(str),
+                   "\n  processing string starting at line %d",
+                   inf->string_start_line);
+    }
+    while ((inf = inf->included_from)) {  /* local pointer assignment */
+      cat_snprintf(str, sizeof(str), "\n  included from file \"%s\", line %d",
+                   inf_filename(inf), inf->line_num);
+    }
   }
+
+  return str;
 }
 
-/********************************************************************** 
-  ...
-***********************************************************************/
-static void inf_warn(struct inputfile *inf, const char *message)
-{
-  inf_log(inf, LOG_NORMAL, message);
-}
-
-/********************************************************************** 
-  ...
+/*******************************************************************//**
+  Returns token of given type from given inputfile.
 ***********************************************************************/
 const char *inf_token(struct inputfile *inf, enum inf_token_type type)
 {
   const char *c;
   const char *name;
   get_token_fn_t func;
-  
-  assert_sanity(inf);
-  assert(type>=INF_TOK_FIRST && type<INF_TOK_LAST);
+
+  if (!inf_sanity_check(inf)) {
+    return NULL;
+  }
+
+  fc_assert_ret_val(INF_TOK_FIRST <= type && INF_TOK_LAST > type, NULL);
 
   name = tok_tab[type].name ? tok_tab[type].name : "(unnamed)";
   func = tok_tab[type].func;
-  
+
   if (!func) {
-    freelog(LOG_ERROR, "token type %d (%s) not supported yet", type, name);
+    log_error("token type %d (%s) not supported yet", type, name);
     c = NULL;
   } else {
-    if (!have_line(inf))
-      (void) read_a_line(inf);
+    while (!have_line(inf) && read_a_line(inf)) {
+      /* Nothing. */
+    }
     if (!have_line(inf)) {
       c = NULL;
     } else {
@@ -601,70 +633,80 @@ const char *inf_token(struct inputfile *inf, enum inf_token_type type)
     }
   }
   if (c && INF_DEBUG_FOUND) {
-    freelog(LOG_DEBUG, "inputfile: found %s '%s'", name, inf->token.str);
+    log_debug("inputfile: found %s '%s'", name, astr_str(&inf->token));
   }
   return c;
 }
 
-/********************************************************************** 
+/*******************************************************************//**
   Read as many tokens of specified type as possible, discarding
   the results; returns number of such tokens read and discarded.
 ***********************************************************************/
 int inf_discard_tokens(struct inputfile *inf, enum inf_token_type type)
 {
   int count = 0;
-  
-  while(inf_token(inf, type))
+
+  while (inf_token(inf, type)) {
     count++;
+  }
+
   return count;
 }
 
-/********************************************************************** 
-  ...
+/*******************************************************************//**
+  Returns section name in current position of inputfile. Returns NULL
+  if there is no section name on that position. Sets inputfile position
+  after section name.
 ***********************************************************************/
 static const char *get_token_section_name(struct inputfile *inf)
 {
-  char *c, *start;
-  
-  assert(have_line(inf));
+  const char *c, *start;
 
-  c = inf->cur_line.str + inf->cur_line_pos;
-  if (*c++ != '[')
+  fc_assert_ret_val(have_line(inf), NULL);
+
+  c = astr_str(&inf->cur_line) + inf->cur_line_pos;
+  if (*c++ != '[') {
     return NULL;
+  }
   start = c;
   while (*c != '\0' && *c != ']') {
     c++;
   }
-  if (*c != ']')
+  if (*c != ']') {
     return NULL;
-  *c++ = '\0';
-  inf->cur_line_pos = c - inf->cur_line.str;
-  astr_minsize(&inf->token, strlen(start)+1);
-  strcpy(inf->token.str, start);
-  return inf->token.str;
+  }
+  *((char *) c) = '\0'; /* Tricky. */
+  astr_set(&inf->token, "%s", start);
+  *((char *) c) = ']';  /* Revert. */
+  inf->cur_line_pos = c + 1 - astr_str(&inf->cur_line);
+  return astr_str(&inf->token);
 }
 
-/********************************************************************** 
-  ...
+/*******************************************************************//**
+  Returns next entry name from inputfile. Skips white spaces and
+  comments. Sets inputfile position after entry name.
 ***********************************************************************/
 static const char *get_token_entry_name(struct inputfile *inf)
 {
-  char *c, *start, *end;
-  
-  assert(have_line(inf));
+  const char *c, *start, *end;
+  char trailing;
 
-  c = inf->cur_line.str + inf->cur_line_pos;
-  while(*c != '\0' && my_isspace(*c)) {
+  fc_assert_ret_val(have_line(inf), NULL);
+
+  c = astr_str(&inf->cur_line) + inf->cur_line_pos;
+  while (*c != '\0' && fc_isspace(*c)) {
     c++;
   }
-  if (*c == '\0')
+  if (*c == '\0') {
     return NULL;
+  }
   start = c;
-  while (*c != '\0' && !my_isspace(*c) && *c != '=' && !is_comment(*c)) {
+  while (*c != '\0' && !fc_isspace(*c) && *c != '=' && !is_comment(*c)) {
     c++;
   }
-  if (!(*c != '\0' && (my_isspace(*c) || *c == '='))) 
+  if (!(*c != '\0' && (fc_isspace(*c) || *c == '='))) {
     return NULL;
+  }
   end = c;
   while (*c != '\0' && *c != '=' && !is_comment(*c)) {
     c++;
@@ -672,145 +714,238 @@ static const char *get_token_entry_name(struct inputfile *inf)
   if (*c != '=') {
     return NULL;
   }
-  *end = '\0';
-  inf->cur_line_pos = c + 1 - inf->cur_line.str;
-  astr_minsize(&inf->token, strlen(start)+1);
-  strcpy(inf->token.str, start);
-  return inf->token.str;
+  trailing = *end;
+  *((char *) end) = '\0';       /* Tricky. */
+  astr_set(&inf->token, "%s", start);
+  *((char *) end) = trailing;   /* Revert. */
+  inf->cur_line_pos = c + 1 - astr_str(&inf->cur_line);
+  return astr_str(&inf->token);
 }
 
-/********************************************************************** 
-  ...
+/*******************************************************************//**
+  If inputfile is at end-of-line, frees current line, and returns " ".
+  If there is still something on that line, returns NULL. 
 ***********************************************************************/
 static const char *get_token_eol(struct inputfile *inf)
 {
-  char *c;
-  
-  assert(have_line(inf));
+  const char *c;
+
+  fc_assert_ret_val(have_line(inf), NULL);
 
   if (!at_eol(inf)) {
-    c = inf->cur_line.str + inf->cur_line_pos;
-    while(*c != '\0' && my_isspace(*c)) {
+    c = astr_str(&inf->cur_line) + inf->cur_line_pos;
+    while (*c != '\0' && fc_isspace(*c)) {
       c++;
     }
-    if (*c != '\0' && !is_comment(*c))
+    if (*c != '\0' && !is_comment(*c)) {
       return NULL;
+    }
   }
 
   /* finished with this line: say that we don't have it any more: */
-  inf->cur_line.n = 0;
+  astr_clear(&inf->cur_line);
   inf->cur_line_pos = 0;
-  
-  assign_flag_token(&inf->token, ' ');
-  return inf->token.str;
+
+  astr_set(&inf->token, " ");
+  return astr_str(&inf->token);
 }
 
-/********************************************************************** 
+/*******************************************************************//**
   Get a flag token of a single character, with optional
-  preceeding whitespace.
+  preceding whitespace.
 ***********************************************************************/
 static const char *get_token_white_char(struct inputfile *inf,
-					char target)
+                                        char target)
 {
-  char *c;
-  
-  assert(have_line(inf));
+  const char *c;
 
-  c = inf->cur_line.str + inf->cur_line_pos;
-  while(*c != '\0' && my_isspace(*c)) {
+  fc_assert_ret_val(have_line(inf), NULL);
+
+  c = astr_str(&inf->cur_line) + inf->cur_line_pos;
+  while (*c != '\0' && fc_isspace(*c)) {
     c++;
   }
-  if (*c != target)
+  if (*c != target) {
     return NULL;
-  inf->cur_line_pos = c + 1 - inf->cur_line.str;
-  assign_flag_token(&inf->token, target);
-  return inf->token.str;
+  }
+  inf->cur_line_pos = c + 1 - astr_str(&inf->cur_line);
+  astr_set(&inf->token, "%c", target);
+  return astr_str(&inf->token);
 }
 
-/********************************************************************** 
-  ...
+/*******************************************************************//**
+  Get flag token for table start, or NULL if that is not next token.
 ***********************************************************************/
 static const char *get_token_table_start(struct inputfile *inf)
 {
   return get_token_white_char(inf, '{');
 }
 
-/********************************************************************** 
-  ...
+/*******************************************************************//**
+  Get flag token for table end, or NULL if that is not next token.
 ***********************************************************************/
 static const char *get_token_table_end(struct inputfile *inf)
 {
   return get_token_white_char(inf, '}');
 }
 
-/********************************************************************** 
-  ...
+/*******************************************************************//**
+  Get flag token comma, or NULL if that is not next token.
 ***********************************************************************/
 static const char *get_token_comma(struct inputfile *inf)
 {
   return get_token_white_char(inf, ',');
 }
 
-/********************************************************************** 
+/*******************************************************************//**
   This one is more complicated; note that it may read in multiple lines.
 ***********************************************************************/
 static const char *get_token_value(struct inputfile *inf)
 {
   struct astring *partial;
-  char *c, *start;
+  const char *c, *start;
   char trailing;
   bool has_i18n_marking = FALSE;
   char border_character = '\"';
-  
-  assert(have_line(inf));
 
-  c = inf->cur_line.str + inf->cur_line_pos;
-  while(*c != '\0' && my_isspace(*c)) {
+  fc_assert_ret_val(have_line(inf), NULL);
+
+  c = astr_str(&inf->cur_line) + inf->cur_line_pos;
+  while (*c != '\0' && fc_isspace(*c)) {
     c++;
   }
-  if (*c == '\0')
+  if (*c == '\0') {
     return NULL;
+  }
 
-  if (*c == '-' || my_isdigit(*c)) {
+  if (*c == '-' || *c == '+' || fc_isdigit(*c)) {
     /* a number: */
     start = c++;
-    while(*c != '\0' && my_isdigit(*c)) {
+    while (*c != '\0' && fc_isdigit(*c)) {
       c++;
     }
+    if (*c == '.') {
+      /* Float maybe */
+      c++;
+      while (*c != '\0' && fc_isdigit(*c)) {
+        c++;
+      }
+    }
     /* check that the trailing stuff is ok: */
-    if (!(*c == '\0' || *c == ',' || my_isspace(*c) || is_comment(*c))) {
+    if (!(*c == '\0' || *c == ',' || fc_isspace(*c) || is_comment(*c))) {
       return NULL;
     }
     /* If its a comma, we don't want to obliterate it permanently,
-     * so rememeber it: */
+     * so remember it: */
     trailing = *c;
-    *c = '\0';
-    
-    inf->cur_line_pos = c - inf->cur_line.str;
-    astr_minsize(&inf->token, strlen(start)+1);
-    strcpy(inf->token.str, start);
-    
-    *c = trailing;
-    return inf->token.str;
+    *((char *) c) = '\0';       /* Tricky. */
+
+    inf->cur_line_pos = c - astr_str(&inf->cur_line);
+    astr_set(&inf->token, "%s", start);
+
+    *((char *) c) = trailing;   /* Revert. */
+    return astr_str(&inf->token);
   }
 
   /* allow gettext marker: */
-  if (*c == '_' && *(c+1) == '(') {
+  if (*c == '_' && *(c + 1) == '(') {
     has_i18n_marking = TRUE;
     c += 2;
-    while(*c != '\0' && my_isspace(*c)) {
+    while (*c != '\0' && fc_isspace(*c)) {
       c++;
     }
-    if (*c == '\0')
+    if (*c == '\0') {
       return NULL;
+    }
   }
 
   border_character = *c;
 
-  if (border_character != '\"'
+  if (border_character == '*') {
+    const char *rfname;
+    fz_FILE *fp;
+    bool eof;
+    int pos;
+
+    c++;
+
+    start = c;
+    while (*c != '*') {
+      if (*c == '\0' || *c == '\n') {
+        return NULL;
+      }
+      c++;
+    }
+    c++;
+    /* check that the trailing stuff is ok: */
+    if (!(*c == '\0' || *c == ',' || fc_isspace(*c) || is_comment(*c))) {
+      return NULL;
+    }
+    /* We don't want to obliterate ending '*' permanently,
+     * so remember it: */
+    trailing = *(c - 1);
+    *((char *) (c - 1)) = '\0';     /* Tricky. */
+
+    rfname = fileinfoname(get_data_dirs(), start);
+    if (rfname == NULL) {
+      inf_log(inf, LOG_ERROR, 
+              _("Cannot find stringfile \"%s\"."), start);
+      *((char *) c) = trailing; /* Revert. */
+      return NULL;
+    }
+    *((char *) c) = trailing; /* Revert. */
+    fp = fz_from_file(rfname, "r", -1, 0);
+    if (!fp) {
+      inf_log(inf, LOG_ERROR,
+              _("Cannot open stringfile \"%s\"."), rfname);
+      return NULL;
+    }
+    log_debug("Stringfile \"%s\" opened ok", start);
+    *((char *) (c - 1)) = trailing; /* Revert. */
+    astr_set(&inf->token, "*"); /* Mark as a string read from a file */
+
+    eof = FALSE;
+    pos = 1; /* Past 'filestring' marker */
+    while (!eof) {
+      char *ret;
+
+      ret = fz_fgets((char *) astr_str(&inf->token) + pos,
+                     astr_capacity(&inf->token) - pos, fp);
+      if (ret == NULL) {
+        eof = TRUE;
+      } else {
+        pos = astr_len(&inf->token);
+        astr_reserve(&inf->token, pos + 200);
+      }
+    }
+
+    fz_fclose(fp);
+
+    inf->cur_line_pos = c + 1 - astr_str(&inf->cur_line);
+
+    return astr_str(&inf->token);
+  } else if (border_character != '\"'
       && border_character != '\''
       && border_character != '$') {
-    return NULL;
+    /* A one-word string: maybe FALSE or TRUE. */
+    start = c;
+    while (fc_isalnum(*c)) {
+      c++;
+    }
+    /* check that the trailing stuff is ok: */
+    if (!(*c == '\0' || *c == ',' || fc_isspace(*c) || is_comment(*c))) {
+      return NULL;
+    }
+    /* If its a comma, we don't want to obliterate it permanently,
+     * so remember it: */
+    trailing = *c;
+    *((char *) c) = '\0';       /* Tricky. */
+
+    inf->cur_line_pos = c - astr_str(&inf->cur_line);
+    astr_set(&inf->token, "%s", start);
+
+    *((char *) c) = trailing;   /* Revert. */
+    return astr_str(&inf->token);
   }
 
   /* From here, we know we have a string, we just have to find the
@@ -830,20 +965,17 @@ static const char *get_token_value(struct inputfile *inf)
   inf->string_start_line = inf->line_num;
   inf->in_string = TRUE;
 
-  partial = &inf->partial;	/* abbreviation */
-  astr_minsize(partial, 1);
-  partial->str[0] = '\0';
-  
-  start = c++;			/* start includes the initial \", to
-				   distinguish from a number */
-  for(;;) {
-    int pos;
-    
-    while(*c != '\0' && *c != border_character) {
+  partial = &inf->partial;      /* abbreviation */
+  astr_clear(partial);
+
+  start = c++;                  /* start includes the initial \", to
+                                 * distinguish from a number */
+  for (;;) {
+    while (*c != '\0' && *c != border_character) {
       /* skip over escaped chars, including backslash-doublequote,
-	 and backslash-backslash: */
-      if (*c == '\\' && *(c+1) != '\0') {  
-	c++;
+       * and backslash-backslash: */
+      if (*c == '\\' && *(c + 1) != '\0') {
+        c++;
       }
       c++;
     }
@@ -853,30 +985,25 @@ static const char *get_token_value(struct inputfile *inf)
       break;
     }
 
-    /* Accumulate to partial string and try more lines;
-     * note partial->n must be _exactly_ the right size, so we
-     * can strcpy instead of strcat-ing all the way to the end
-     * each time. */
-    pos = partial->n - 1;
-    astr_minsize(partial, partial->n + c - start + 1);
-    strcpy(partial->str + pos, start);
-    strcpy(partial->str + partial->n - 2, "\n");
-    
+    astr_add(partial, "%s\n", start);
+
     if (!read_a_line(inf)) {
       /* shouldn't happen */
       inf_log(inf, LOG_ERROR, 
               "Bad return for multi-line string from read_a_line");
       return NULL;
     }
-    c = start = inf->cur_line.str;
+    c = start = astr_str(&inf->cur_line);
   }
 
   /* found end of string */
-  *c = '\0';
-  inf->cur_line_pos = c + 1 - inf->cur_line.str;
-  astr_minsize(&inf->token, partial->n + strlen(start));
-  strcpy(inf->token.str, partial->str);
-  strcpy(inf->token.str + partial->n - 1, start);
+  trailing = *c;
+  *((char *) c) = '\0';         /* Tricky. */
+
+  inf->cur_line_pos = c + 1 - astr_str(&inf->cur_line);
+  astr_set(&inf->token, "%s%s", astr_str(partial), start);
+
+  *((char *) c) = trailing;     /* Revert. */
 
   /* check gettext tag at end: */
   if (has_i18n_marking) {
@@ -887,5 +1014,5 @@ static const char *get_token_value(struct inputfile *inf)
     }
   }
   inf->in_string = FALSE;
-  return inf->token.str;
+  return astr_str(&inf->token);
 }
